/* gfx.c */



#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include "support.h"
#include "puzzle.h"
#include "interface.h"
#include "gfx.h"
#include "prefs.h"


#define FNAME_OVERLAY_BLACK PACKAGE_PIXMAP_DIR G_DIR_SEPARATOR_S "overlayblack.png"
#define FNAME_OVERLAY_WHITE PACKAGE_PIXMAP_DIR G_DIR_SEPARATOR_S "overlaywhite.png"
#define FNAME_OVERLAY_EMPTY PACKAGE_PIXMAP_DIR G_DIR_SEPARATOR_S "overlayempty.png"


extern Puzzle    *puzzle;
extern GtkWidget *app;
extern GtkWidget *drawarea;
extern Overlay   overlay;
extern gint      img_id;
extern gchar     *fname_img_custom;
extern ImgFile   img_file[];
extern gint      n_img_files;
extern gint      image_px;

GdkPixmap *pixmap_display = NULL;  /* image of current game state */

static GdkPixbuf *pixbuf_image   = NULL;  /* original image scaled to IMAGE_WIDTH x IMAGE_HEIGHT */
static GdkPixmap *pixmap_tileset = NULL;  /* copy of pixbuf_image with overlays drawn on top     */
static GdkPixmap *pixmap_missing = NULL;  /* the "missing" tile, with a shaded overlay           */

static gint tile_px;
static gint x_blink;
static gint y_blink;
static gint n_blink;
static guint timeout_id;



void
gfx_free (void)
{
	if (pixbuf_image != NULL) g_object_unref (pixbuf_image);
	if (pixmap_tileset != NULL) g_object_unref (pixmap_tileset);
	if (pixmap_missing != NULL) g_object_unref (pixmap_missing);
}



void
gfx_draw_tile (gint x, gint y, gboolean refresh, gboolean blink)
{
        Tile *t = puzzle_get_tile (puzzle, x, y);

        if (IS_MISSING(t) && (!IS_SOLVED(puzzle) || blink))  {
                /* draw as missing tile */
                gdk_draw_drawable (pixmap_display,
                                   drawarea->style->black_gc,
                                   pixmap_missing,
                                   0, 0,
                                   x * tile_px, y * tile_px,
                                   tile_px, tile_px);
        }
        else {
                /* draw as normal tile */
                gdk_draw_drawable (pixmap_display,
                                   drawarea->style->black_gc,
                                   pixmap_tileset,
                                   tile_get_x (t), tile_get_y (t),
                                   x * tile_px, y * tile_px,
                                   tile_px, tile_px);
        }

        if (refresh) {
                gtk_widget_queue_draw_area (drawarea, 0, 0, image_px, image_px);
        }
}



void
gfx_redraw_drawarea (gboolean refresh)
{
        gint x, y, size = puzzle_get_size (puzzle);

        /* g_printerr ("gfx_redraw_drawarea\n"); */

        for (y = 0; y < size; y++) {
                for (x = 0; x < size; x++) {
                        gfx_draw_tile (x, y, FALSE, FALSE);
                }
        }
        if (refresh) {
                gtk_widget_queue_draw_area (drawarea, 0, 0, image_px, image_px);
        }
}



void
gfx_expose (GdkRectangle *area)
{
        gdk_draw_drawable (drawarea->window,
                           app->style->fg_gc[GTK_WIDGET_STATE(drawarea)],
                           pixmap_display,
                           area->x, area->y,
                           area->x, area->y,
                           area->width, area->height);
}



static gint
on_gfx_blink_tile (Tile *data)
{
        Tile *t = data;

        gfx_draw_tile (x_blink, y_blink, TRUE, TRUE);

        tile_set_missing (t, !IS_MISSING(t));
        n_blink++;

        if (n_blink > 6 && IS_MISSING(t)) {
                timeout_id = 0;
                return FALSE;
        }
        return TRUE;
}



void
gfx_kill_timeout (void)
{
        /* called when quitting program */
        if (timeout_id != 0) {
                gtk_timeout_remove (timeout_id);
                timeout_id = 0;
        }
}



void
gfx_blink_tile (gint x, gint y)
{
        Tile *t = puzzle_get_tile (puzzle, x, y);

        x_blink = x;
        y_blink = y;
        n_blink = 0;
        timeout_id = gtk_timeout_add (250, (GtkFunction)on_gfx_blink_tile, (gpointer)t);
}



static void
gfx_create_missing_tile (void)
{
        GdkPixbuf *pixbuf_tmp;          /* overlay */
        GdkPixbuf *pixbuf_tmp_scaled;   /* overlay scaled to tile size */
        GdkPixbuf *pixbuf_image_scaled; /* puzzle image scaled to tile size */


	pixmap_missing = gdk_pixmap_new (app->window, tile_px, tile_px, -1);
        pixbuf_image_scaled = gdk_pixbuf_scale_simple (pixbuf_image, tile_px, tile_px, GDK_INTERP_BILINEAR);
        gdk_pixbuf_render_pixmap_and_mask (pixbuf_image_scaled, &pixmap_missing, NULL, 127);

        pixbuf_tmp = gdk_pixbuf_new_from_file (FNAME_OVERLAY_EMPTY, NULL);
        if (pixbuf_tmp != NULL) {
                pixbuf_tmp_scaled = gdk_pixbuf_scale_simple (pixbuf_tmp,
                                                             tile_px, tile_px,
                                                             GDK_INTERP_BILINEAR);
                gdk_pixbuf_render_to_drawable_alpha
                        (pixbuf_tmp_scaled, pixmap_missing, 0, 0, 0, 0, tile_px, tile_px,
                         GDK_PIXBUF_ALPHA_FULL, 127, GDK_RGB_DITHER_NONE, 0, 0);
                g_object_unref (pixbuf_tmp);
                g_object_unref (pixbuf_tmp_scaled);

        }
        else {
                g_printerr (PACKAGE ": can't load image: %s\n", FNAME_OVERLAY_EMPTY);
                gdk_pixbuf_render_to_drawable_alpha
                        (pixbuf_image_scaled, pixmap_missing, 0, 0, 0, 0, tile_px, tile_px,
                         GDK_PIXBUF_ALPHA_FULL, 127, GDK_RGB_DITHER_NONE, 0, 0);
        }
        g_object_unref (pixbuf_image_scaled);
}



void
gfx_create_tileset (void)
{
        GdkPixbuf *pixbuf_tmp;
        GdkPixbuf *pixbuf_tmp_scaled;
        gchar *fname_overlay = NULL;
        gint x, y, size;


        /* g_printerr ("gfx_create_tileset\n"); */
        
        size = puzzle_get_size (puzzle);
        tile_px = puzzle_get_tile_px (puzzle);
        gdk_pixbuf_render_pixmap_and_mask (pixbuf_image,
                &pixmap_tileset, NULL, 127);

	switch(overlay) {
	case OVERLAY_BLACK :
		fname_overlay = FNAME_OVERLAY_BLACK;
		break;
	case OVERLAY_WHITE :
		fname_overlay = FNAME_OVERLAY_WHITE;
		break;
	case OVERLAY_NONE :
		gfx_create_missing_tile ();
		return;
	}


        pixbuf_tmp = gdk_pixbuf_new_from_file (fname_overlay, NULL);
        if (pixbuf_tmp == NULL) {
                g_printerr (PACKAGE ": can't load image: %s\n", fname_overlay);
                gfx_create_missing_tile ();
                return;
        }
        pixbuf_tmp_scaled = gdk_pixbuf_scale_simple (pixbuf_tmp,
                tile_px, tile_px, GDK_INTERP_BILINEAR);

        for (y = 0; y < size; y++) {
                for (x = 0; x < size; x++) {
                        gdk_pixbuf_render_to_drawable_alpha
                                (pixbuf_tmp_scaled, pixmap_tileset,
                                 0, 0,
                                 x * tile_px, y * tile_px,
                                 tile_px, tile_px,
                                 GDK_PIXBUF_ALPHA_FULL, 127,
                                 GDK_RGB_DITHER_NONE, 0, 0);
                }
        }

        g_object_unref (pixbuf_tmp);
        g_object_unref (pixbuf_tmp_scaled);

        gfx_create_missing_tile ();
}



gboolean
gfx_load (const gchar *fname, gboolean was_drag_n_drop)
{
        /* Load the named image, create a tile set from it,
         * and update the display. Return TRUE if successful.
         */

        GdkPixbuf *pixbuf_tmp;


        pixbuf_tmp = gdk_pixbuf_new_from_file (fname, NULL);
        if (pixbuf_tmp == NULL) {
                if (!was_drag_n_drop) {
                        g_printerr (PACKAGE ": can't load image: %s\n", fname);
                }
                return FALSE;
       }

        gfx_free (); /* free old graphics now */

	pixbuf_image = gdk_pixbuf_scale_simple (pixbuf_tmp,
                                                image_px, image_px,
                                                GDK_INTERP_BILINEAR);
        g_object_unref (pixbuf_tmp);

	gfx_create_tileset ();
	gfx_redraw_drawarea (TRUE);

        return TRUE;
}



gboolean
gfx_load_img_id (gint id)
{
        /* Load and display the image specified by id.
         * Save as new preference.
         * Return TRUE if successful.
         */

        gchar *fname = NULL;


        if (id == IMG_ID_CUSTOM) {
                fname = fname_img_custom;
        }
        else {
                fname = img_file[id].fname;
        }

        if (img_file[id].size != 0) {
                puzzle_set_size (puzzle, img_file[id].size);
        }

        if (!gfx_load (fname, FALSE)) return FALSE;

        if (id != img_id) prefs_set_image_id (id);
        return TRUE;
}

