/* interface.c */



#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <gconf/gconf-client.h>

#include "support.h"
#include "string.h"
#include "puzzle.h"
#include "interface.h"
#include "gfx.h"
#include "prefs.h"



extern Puzzle    *puzzle;
extern GdkPixmap *pixmap_display;
extern gchar     *fname_img_custom;
extern KeyID     key[4];
extern gint      img_id;
extern ImgFile   img_file[];
extern gint      n_img_files;
extern GConfClient *skoosh_gconf_client;
extern gint      image_px;


typedef enum {
        SOUND_MOVE,
        SOUND_SOLVED
} SoundID;


GtkWidget *app;
GtkWidget *drawarea;
GtkWidget *appbar;
gchar     *fname_icon = NULL;



static GtkTargetEntry target_table[] = {
        { "text/uri-list", 0, 0 },
        { "text/plain", 0, 1 }
};



static void
prog_cleanup (void)
{
        gfx_kill_timeout ();

        gtk_main_quit ();

        gfx_free ();
        g_free (fname_icon);
        puzzle_free (puzzle);
        prefs_free ();
}



static void
play_sound (SoundID s)
{
        switch (s) {
            case SOUND_MOVE:
                gnome_triggers_do (NULL, NULL, PACKAGE, "move", NULL);
                break;
            case SOUND_SOLVED:
                gnome_triggers_do (NULL, NULL, PACKAGE, "solve", NULL);
                break;
        }
}



static void
create_dialog_about (void)
{
        static GtkWidget *dialog_about = NULL;
        const gchar *authors[] = {
                N_("Timothy Musson <trmusson@ihug.co.nz>"),
                N_("\nArtwork Credits:"),
                N_("\n\"Summer Breeze (Hot Mix)\"\n    Copyright 1998, Gregory J. Griffin\n    <griffart@comcast.net>\n    http://mywebpages.comcast.net/gjg13/"),
                NULL
        };
        const gchar *documenters[] = { NULL };
        gchar *translators = _("translator_credits");
        GdkPixbuf *pixbuf_icon = NULL;
        gint i;


        if (dialog_about != NULL) {
                gtk_window_present (GTK_WINDOW(dialog_about));
                return;
        }

        i = 0;
        while (authors[i] != NULL) {
                authors[i] = _(authors[i]);
                i++;
        }

        if (!strcmp (translators, "translator_credits")) {
                translators = NULL;
        }


        if (fname_icon != NULL) {
                pixbuf_icon = gdk_pixbuf_new_from_file (fname_icon, NULL);
        }

        dialog_about = gnome_about_new (PACKAGE, VERSION,
                                        "Copyright 2001-2003, Timothy Musson",
                                        _("Slide the tiles to reveal a picture"),
                                        authors,
                                        documenters,
                                        translators,
                                        pixbuf_icon);

        gtk_window_set_transient_for (GTK_WINDOW (dialog_about), GTK_WINDOW(app));
        g_signal_connect (G_OBJECT(dialog_about), "destroy",
                          G_CALLBACK(gtk_widget_destroyed), &dialog_about);

        if (pixbuf_icon != NULL) g_object_unref (pixbuf_icon);

        gtk_widget_show (dialog_about);
}



static void
on_game_new_activate (GtkMenuItem *menuitem, gpointer user_data)
{
        puzzle_new_game (puzzle);
        set_status (STATUS_MOVES, NULL);
}



static void
on_game_quit_activate (GtkMenuItem *menuitem, gpointer user_data)
{
        prog_cleanup ();
}



static void
on_settings_preferences_activate (GtkMenuItem *menuitem, gpointer user_data)
{
        create_prefs_dialog ();
}



static void
on_help_about_activate (GtkMenuItem *menuitem, gpointer user_data)
{
        create_dialog_about ();
}



static GnomeUIInfo game_menu_uiinfo[] =
{
        GNOMEUIINFO_MENU_NEW_GAME_ITEM (on_game_new_activate, NULL),
        GNOMEUIINFO_SEPARATOR,
        GNOMEUIINFO_MENU_EXIT_ITEM (on_game_quit_activate, NULL),
        GNOMEUIINFO_END
};



static GnomeUIInfo settings_menu_uiinfo[] =
{
        GNOMEUIINFO_MENU_PREFERENCES_ITEM (on_settings_preferences_activate, NULL),
        GNOMEUIINFO_END
};



static GnomeUIInfo help_menu_uiinfo[] =
{
        GNOMEUIINFO_HELP (PACKAGE),
        GNOMEUIINFO_MENU_ABOUT_ITEM (on_help_about_activate, NULL),
        GNOMEUIINFO_END
};



static GnomeUIInfo menubar_uiinfo[] =
{
        GNOMEUIINFO_MENU_GAME_TREE (game_menu_uiinfo),
        GNOMEUIINFO_MENU_SETTINGS_TREE (settings_menu_uiinfo),
        GNOMEUIINFO_MENU_HELP_TREE (help_menu_uiinfo),
        GNOMEUIINFO_END
};



static gboolean
process_move (gint x, gint y)
{
        if (puzzle_move_tile (puzzle, x, y)) {
                puzzle_incr_moves (puzzle);
                set_status (STATUS_MOVES, NULL);
                if (IS_SOLVED(puzzle)) {
                        set_status (STATUS_SOLVED, NULL);
                        play_sound (SOUND_SOLVED);
                        gfx_draw_tile (x, y, TRUE, FALSE);
                        gfx_blink_tile (x, y);
                }
                else {
                    play_sound (SOUND_MOVE);
                }
                return TRUE;
        }
        return FALSE;
}



static void
on_drawarea_event (GtkWidget *widget, const GdkEvent *event)
{
	GdkEventExpose *expose;
	gint x, y;


	switch (event->type) {

	case GDK_EXPOSE :
		expose = (GdkEventExpose *)event;
		gfx_expose (&expose->area);
		gdk_flush ();
		break;

	case GDK_BUTTON_PRESS :
                if (!IS_SOLVED(puzzle)) {
                        gtk_widget_get_pointer (widget, &x, &y);
                        process_move (x / puzzle_get_tile_px (puzzle),
                                      y / puzzle_get_tile_px (puzzle));
                }
		break;

	default:
		break;

	}
}



static gboolean
on_key_press (GtkWidget* widget, GdkEventKey* event, gpointer data)
{
        gint x = 0, y = 0, keyval;

        if (IS_SOLVED(puzzle) == TRUE) return FALSE;

        while (!IS_MISSING(puzzle_get_tile (puzzle, x, y))) {
                x++;
                if (x == puzzle_get_size (puzzle)) {
                        x = 0;
                        y++;
                }
        }

        keyval = event->keyval;
        if (keyval == key[KEY_UP] && y < puzzle_get_size (puzzle) - 1) {
                return process_move (x, y + 1);
        }
        else if (keyval == key[KEY_DN] && y > 0) {
                return process_move (x, y - 1);
        }
        else if (keyval == key[KEY_LT] && x < puzzle_get_size (puzzle) - 1) {
                return process_move (x + 1, y);
        }
        else if (keyval == key[KEY_RT] && x > 0) {
                return process_move (x - 1, y);
        }

        return FALSE;
}



static gboolean
on_drag_received (GtkWidget *widget, GdkDragContext *context,
                  gint x, gint y, GtkSelectionData *data,
                  guint info, guint time)
{
        gchar *fname;
        gint i;

        if (data == NULL) return TRUE;
        if (strncmp (data->data, "file:///", 8) != 0) return FALSE;

        /* g_printerr ("%s\n", data->data); */

        fname = g_strdup (&data->data[7]);
        for (i = 0; i < strlen (fname); i++) {
                if (fname[i] == 13) fname[i] = '\0';
        }

        /* g_printerr ("%s\n", fname); */

        if (gfx_load (fname, TRUE)) {
                prefs_set_image (fname);
                prefs_set_image_id (IMG_ID_CUSTOM);
        }
        else {
                g_free (fname);
        }
        return TRUE;
}



void
set_status (StatusMsgType msgtype, const gchar *msg)
{
        gchar *moves = NULL;

        switch (msgtype) {
        case STATUS_FLASH:
                gnome_app_flash (GNOME_APP(app), msg);
                return;
        case STATUS_MOVES:
                moves = g_strdup_printf (_("Moves: %d"), puzzle_get_moves (puzzle));
                break;
        case STATUS_SOLVED:
                moves = g_strdup_printf (_("Solved in %d moves!"), puzzle_get_moves (puzzle));
                break;
        case STATUS_CLEAR:
                gnome_appbar_clear_stack (GNOME_APPBAR(appbar));
                gnome_appbar_refresh (GNOME_APPBAR(appbar));
                return;
        }
        gnome_appbar_pop (GNOME_APPBAR(appbar));
        gnome_appbar_push (GNOME_APPBAR(appbar), moves);
        g_free (moves);
}



void
file_error_dialog (const gchar *fname, GtkMessageType type)
{
        GtkWidget *dialog;

        dialog = gtk_message_dialog_new (GTK_WINDOW(app),
                                         GTK_DIALOG_MODAL,
                                         type,
                                         GTK_BUTTONS_CLOSE,
                                         _("Unable to load image:\n%s"), fname);

        gtk_dialog_run (GTK_DIALOG(dialog));
        gtk_widget_destroy (dialog);
}



void
create_app (void)
{
        GtkWidget *bonobodock;


        app = gnome_app_new (PACKAGE, "Skoosh");

        g_signal_connect (GTK_OBJECT(app), "destroy", GTK_SIGNAL_FUNC(on_game_quit_activate), NULL);

        fname_icon = gnome_program_locate_file (NULL, GNOME_FILE_DOMAIN_APP_PIXMAP,
                                                ("skoosh.png"),
                                                TRUE, NULL);
        if (fname_icon != NULL) {
                gnome_window_icon_set_default_from_file (fname_icon);
                gnome_window_icon_set_from_default (GTK_WINDOW(app));
        }

        bonobodock = GNOME_APP (app)->dock;
        gtk_widget_show (bonobodock);

        gnome_app_create_menus (GNOME_APP (app), menubar_uiinfo);

        drawarea = gtk_drawing_area_new ();
        gtk_widget_set_size_request (drawarea, image_px, image_px);
        gtk_widget_show (drawarea);
        gnome_app_set_contents (GNOME_APP (app), drawarea);

	gtk_widget_set_events (drawarea, GDK_EXPOSURE_MASK | GDK_BUTTON_PRESS_MASK);
	g_signal_connect (GTK_OBJECT(drawarea), "event", GTK_SIGNAL_FUNC(on_drawarea_event), NULL);

        g_signal_connect (GTK_OBJECT(app), "key_press_event",
                          GTK_SIGNAL_FUNC(on_key_press), NULL);

        gtk_drag_dest_set (drawarea, GTK_DEST_DEFAULT_ALL, target_table, 1, GDK_ACTION_COPY);
        g_signal_connect (GTK_OBJECT(drawarea), "drag_data_received",
                          GTK_SIGNAL_FUNC(on_drag_received), NULL);

        appbar = gnome_appbar_new (FALSE, TRUE, GNOME_PREFERENCES_NEVER);
        gtk_widget_show (appbar);
        gnome_app_set_statusbar (GNOME_APP (app), appbar);

        gnome_app_install_menu_hints (GNOME_APP (app), menubar_uiinfo);

	gtk_widget_realize (drawarea);
	pixmap_display = gdk_pixmap_new (app->window, image_px, image_px, -1);

        gtk_window_set_resizable (GTK_WINDOW(app), FALSE);

        gtk_widget_show (app);

        /* load player's preferred image, trying the default on failure */
        if (!gfx_load_img_id (img_id)) {
                if (img_id == IMG_ID_DEFAULT) {
                        /* default was the preference and it failed */
                        file_error_dialog (img_file[img_id].fname, GTK_MESSAGE_ERROR);
                        exit (1);
                }
                if (img_id == IMG_ID_CUSTOM) {
                        file_error_dialog (fname_img_custom, GTK_MESSAGE_WARNING);
                        g_free (fname_img_custom);
                        fname_img_custom = NULL;
                        gconf_client_unset (skoosh_gconf_client, "/apps/skoosh/image", NULL);
                }
                else {
                        file_error_dialog (img_file[img_id].fname, GTK_MESSAGE_WARNING);
                }
                if (!gfx_load_img_id (IMG_ID_DEFAULT)) {
                        /* default failed */
                        file_error_dialog (img_file[IMG_ID_DEFAULT].fname, GTK_MESSAGE_ERROR);
                        exit (1);
                }
        }

        set_status (STATUS_FLASH, _("Welcome to Skoosh!"));
}

