/* puzzle.c */



#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <stdlib.h>
#include <time.h>

#include "support.h"
#include "puzzle.h"
#include "gfx.h"


extern gint image_px;


#define RANDOM_INT(n)    ((int)((double)rand() / ((double)RAND_MAX + 1) * n)) /* 0..n-1 */
#define N_TILES          (p->size * p->size)
#define ID_LAST_TILE     (N_TILES - 1)
#define TILE(p,x,y)      ((x) + (y) * p->size)
#define MOVE_NO          0
#define MOVE_UP          1
#define MOVE_DN          2
#define MOVE_LT          4
#define MOVE_RT          8



struct _Puzzle {
        Tile     **tile;
        gint     tile_px; /* tile size in pixels (image_px / size) */
        gint     size;    /* number of tiles across and down */
        gint     moves;
        gboolean random;
        gboolean solved;
};


struct _Tile {
        gint     id;
        gint     x;       /* pixel offset in source image */
        gint     y;       /* pixel offset in source image */
        gboolean missing;
};


static void puzzle_free_tiles (Puzzle *p);
static void puzzle_swap_tiles (Puzzle *p, gint i1, gint i2);
static void puzzle_pick_missing_tile (Puzzle *p);
static gint puzzle_get_move_direction (Puzzle *p, gint x, gint y);
static gint puzzle_get_shuffle_options (Puzzle *p, gint x, gint y);
static void puzzle_set_tile_data (Puzzle *p, gint id, gint x, gint y);



Puzzle*
puzzle_new (void)
{
        Puzzle *p = g_new0 (Puzzle, 1);
        puzzle_set_solved (p, TRUE);
        return p;
}

  
  
void
puzzle_free (Puzzle *p)
{
        if (p == NULL) return;
        puzzle_free_tiles (p);
        g_free (p);
}



void
puzzle_new_game (Puzzle *p)
{
        puzzle_shuffle_tiles (p);
        puzzle_clear_moves (p);
        puzzle_set_solved (p, FALSE);
        puzzle_display_tiles (p);
}



void
puzzle_set_size (Puzzle *p, gint size)
{
        /* reset puzzle to size x size tiles */

        puzzle_set_solved (p, TRUE);
        puzzle_free_tiles (p);
        p->size = size;
        p->tile_px = image_px / size; /* tile size in pixels */
        puzzle_create_tiles (p);
}



gint
puzzle_get_size (Puzzle *p)
{
        return p->size;
}



gint
puzzle_get_tile_px (Puzzle *p)
{
        /* return tile size in pixels */
        return p->tile_px;
}



void
puzzle_check_state (Puzzle *p)
{
        /* set p->solved according to tile order */
        gint i;

        for (i = 0; i < N_TILES; i++) {
                if (i != p->tile[i]->id) {
                        puzzle_set_solved (p, FALSE);
                        return;
                }
        }
        puzzle_set_solved (p, TRUE);
}



void
puzzle_set_solved (Puzzle *p, gboolean solved)
{
        p->solved = solved;
}



gboolean
puzzle_get_solved (Puzzle *p)
{
        return p->solved;
}



void
puzzle_clear_moves (Puzzle *p)
{
        p->moves = 0;
}



void
puzzle_incr_moves (Puzzle *p)
{
        p->moves++;
}



gint
puzzle_get_moves (Puzzle *p)
{
        return p->moves;
}



void
puzzle_set_random (Puzzle *p, gboolean random)
{
        p->random = random;
}



gboolean
puzzle_get_random (Puzzle *p)
{
        return p->random;
}



Tile*
puzzle_get_tile (Puzzle *p, gint x, gint y)
{
        return p->tile[TILE(p, x, y)];
}



gboolean
tile_get_missing (Tile *t)
{
        /* return TRUE if t is the missing tile */
        return t->missing;
}

  
  
void
tile_set_missing (Tile *t, gboolean missing)
{
        t->missing = missing;
}



gint
tile_get_x (Tile *t)
{
        /* return x offset for this tile in the source image */
        return t->x;
}



gint
tile_get_y (Tile *t)
{
        /* return y offset for this tile in the source image */
        return t->y;
}



static void
puzzle_swap_tiles (Puzzle *p, gint i1, gint i2)
{
        Tile *tmp = p->tile[i1];

        p->tile[i1] = p->tile[i2];
        p->tile[i2] = tmp;
}



static void
puzzle_pick_missing_tile (Puzzle *p)
{
        /* make one of the tiles the "missing" tile */
  
        gint i;

        for (i = 0; i < N_TILES; i++) p->tile[i]->missing = FALSE;

        if (IS_RANDOM(p)) { /* pick any tile */
                i = RANDOM_INT(N_TILES);
        }
        else {              /* pick the lower-right tile */
                for (i = 0; i < N_TILES; i++) {
                        if (p->tile[i]->id == ID_LAST_TILE) break;
                }
        }
        tile_set_missing (p->tile[i], TRUE);
}



static void
puzzle_free_tiles (Puzzle *p)
{
        gint i;

        if (p->tile == NULL) return;
        for (i = 0; i < N_TILES; i++) g_free (p->tile[i]);
        g_free (p->tile);
        p->tile = NULL;
}



static void
puzzle_set_tile_data (Puzzle *p, gint id, gint x, gint y)
{
        p->tile[id]->id = id;
        p->tile[id]->x  = p->tile_px * x;
        p->tile[id]->y  = p->tile_px * y;
}



void
puzzle_create_tiles (Puzzle *p)
{
        gint x, y, id = 0;

        if (p->tile != NULL) puzzle_free_tiles (p);
        p->tile = g_malloc (sizeof (Tile*) * N_TILES);
        for (y = 0; y < p->size; y++) {
                for (x = 0; x < p->size; x++) {
                        p->tile[id] = g_new (Tile, 1);
                        puzzle_set_tile_data (p, id, x, y);
                        id++;
                }
        }
}



void
puzzle_reset_tiles (Puzzle *p)
{
        gint x, y, id = 0;

        for (y = 0; y < p->size; y++) {
                for (x = 0; x < p->size; x++) {
                        puzzle_set_tile_data (p, id, x, y);
                        id++;
                }
        }
        puzzle_pick_missing_tile (p);
}



void
puzzle_display_tiles (Puzzle *p)
{
        gfx_redraw_drawarea (TRUE);
}



static gint
puzzle_get_move_direction (Puzzle *p, gint x, gint y)
{
        /* the tile at x y can... */
	if (y > 0 && puzzle_get_tile (p, x, y - 1)->missing)
                return MOVE_UP;
	if (y < p->size - 1 && puzzle_get_tile (p, x, y + 1)->missing)
                return MOVE_DN;
        if (x > 0 && puzzle_get_tile (p, x - 1, y)->missing)
                return MOVE_LT;
	if (x < p->size - 1 && puzzle_get_tile (p, x + 1, y)->missing)
                return MOVE_RT;
	return MOVE_NO;
}



gboolean
puzzle_move_tile (Puzzle *p, gint x, gint y)
{
        /* try to move the tile that's at x y */
        switch (puzzle_get_move_direction (p, x, y)) {
        case MOVE_UP:
                puzzle_swap_tiles (p, TILE(p, x, y), TILE(p, x, y - 1));
                gfx_draw_tile (x, y - 1, FALSE, FALSE);
                break;
        case MOVE_DN:
                puzzle_swap_tiles (p, TILE(p, x, y), TILE(p, x, y + 1));
                gfx_draw_tile (x, y + 1, FALSE, FALSE);
                break;
        case MOVE_LT:
                puzzle_swap_tiles (p, TILE(p, x, y), TILE(p, x - 1, y));
                gfx_draw_tile (x - 1, y, FALSE, FALSE);
                break;
        case MOVE_RT:
                puzzle_swap_tiles (p, TILE(p, x, y), TILE(p, x + 1, y));
                gfx_draw_tile (x + 1, y, FALSE, FALSE);
                break;
        case MOVE_NO:
                return FALSE;
        }
        gfx_draw_tile (x, y, TRUE, FALSE);
        /* puzzle_display_tiles (p); */
        puzzle_check_state (p);
        return TRUE;
}



static gint
puzzle_get_shuffle_options (Puzzle *p, gint x, gint y)
{
        gint options = MOVE_NO;

        /* given the missing tile's x y position... */
	if (y > 0)
                options = MOVE_DN;      /* the tile above can move down    */
	if (y < p->size - 1)
                options |= MOVE_UP;     /* the tile below can move up      */
	if (x > 0)
                options |= MOVE_RT;     /* the tile to left can move right */
	if (x < p->size - 1)
                options |= MOVE_LT;     /* the tile to right can move left */

	return options;
}



void
puzzle_shuffle_tiles (Puzzle *p)
{
        gint i, options, x = 0, y = 0;


        puzzle_reset_tiles (p);

        /* get the missing tile's x y position */
        for (i = 0; i < N_TILES; i++) {
                if (IS_MISSING(p->tile[i])) break;
                x++;
                if (x == p->size) {
                        x = 0;
                        y++;
                }
        }


        do {

                for (i = 0; i < 100 * N_TILES; i++) {

                        options = puzzle_get_shuffle_options (p, x, y);

                        if (options & MOVE_UP && RANDOM_INT(4) == 1) {
                                puzzle_swap_tiles (p, TILE(p, x, y + 1), TILE(p, x, y));
                                y++;
                        }
                        else if (options & MOVE_DN && RANDOM_INT(4) == 1) {
                                puzzle_swap_tiles (p, TILE(p, x, y - 1), TILE(p, x, y));
                                y--;
                        }
                        else if (options & MOVE_LT && RANDOM_INT(4) == 1) {
                                puzzle_swap_tiles (p, TILE(p, x + 1, y), TILE(p, x, y));
                                x++;
                        }
                        else if (options & MOVE_RT && RANDOM_INT(4) == 1) {
                                puzzle_swap_tiles (p, TILE(p, x - 1, y), TILE(p, x, y));
                                x--;
                        }
                }

                puzzle_check_state (p);

        } while (IS_SOLVED(p));
}


